from __future__ import annotations

import numpy as np

from .answer import Answer


class Question:
    """A class to represent a Question on a form or template."""

    def __init__(
        self, question_id: int, answers: list[Answer], multiple: bool
    ):
        """Return Question instance to represent group of answers belonging to a question.

        Args:
            question_id (int): Unique ID to keep track of question number.
            form_img (np.ndarray): image of form.
            answers (list[Answer]): All spots that reference a possible answer.
            multiple (bool): Set to true if question has multiple possible answers.
        """
        self.answers = answers
        self.multiple = multiple
        self.question_id = question_id

    @property
    def question_img(self, form_img: np.ndarray) -> np.ndarray:
        """return cropped form image of the question

        Args:
            form_img (np.ndarray): form image read into array e.g. via cv2.imread()

        Returns:
            np.ndarray: cropped form image
        """
        return form_img[
            self.__search_area_y0 : self.__search_area_y1,
            self.__search_area_x0 : self.__search_area_x1,
        ]

    @property
    def __search_area_x0(self):
        return min([ans.x for ans in self.answers])

    @property
    def __search_area_x1(self):
        return max([ans.x for ans in self.answers])

    @property
    def __search_area_y0(self):
        return min([ans.y for ans in self.answers])

    @property
    def __search_area_y1(self):
        return max([ans.y for ans in self.answers])

    def find_answers(self, img: np.ndarray) -> list[Answer]:
        """Find marked answer(s) for question

        Args:
            img (np.ndarray): image of the form

        Returns:
            list[Answer]: answers that have been marked i.e. return true for answer.check_fill()
            if question.multiple == False then the first marked answer will be returned and length
            of this list will be 1.
        """

        answers = []
        for ans in self.answers:
            if ans.is_filled(img):
                if self.multiple:
                    answers.append(ans)
                else:
                    return [ans]
        return answers



from __future__ import annotations

import json
from typing import TYPE_CHECKING

import cv2
import numpy as np

import formpy.utils.img_processing as ip
from formpy.answer import Answer
from formpy.question import Question
from formpy.utils.template_definition import find_spots

if TYPE_CHECKING:
    from formpy.form import Form


class Template:
    """A class to represent a template that a form is built from."""

    def __init__(
        self, img: np.ndarray, questions: list[Question], circle_radius: int
    ):
        """initialise template

        Args:
            img (np.ndarray): image of template read in using e.g. cv2.imread()
            questions (list[Question]): list of questions on template
            circle_radius (int): size of answer circles
        """
        self.questions = questions
        self.img = ip.process_img(img)
        self.circle_radius = circle_radius

    @classmethod
    def from_img_template(
        cls,
        img_path: str,
        circle_radius: int,
        question_assignment: dict,
        question_config: dict = None,
    ) -> Template:
        """Initialise template from img

        Args:
            img_path (str): path to load image of template from
            circle_radius (int): size of the answer circles
            question_assignment (dict): map of question id to list of answer ids.
            question_config (dict, optional): map of question id to true/false flag for multiple answers. Defaults to None.

        Returns:
            Template
        """

        # load image and align
        raw_img = cv2.imread(img_path)
        img = ip.process_img(raw_img)

        # find all spots - sorted by x then y
        all_spots = find_spots(
            img, max_radius=circle_radius + 5, min_radius=circle_radius - 5
        )
        unassigned_answers = all_spots.copy()

        # create colour image to highlight answers in red
        colour_img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)

        # initialise question_id
        question_id = 0
        # list to hold assigned questions
        questions = []
        # keep looping until all found spots are assigned to a questions
        while len(unassigned_answers) > 0:
            # list to hold assigned answers
            assigned_answers = []
            for i, spot in enumerate(all_spots):
                # color unassigned qns in red and assigned in green
                if spot in unassigned_answers:
                    color = (0, 0, 255)
                else:
                    color = (0, 255, 0)
                cv2.putText(
                    colour_img,
                    str(i),
                    (spot[0], spot[1]),
                    cv2.FONT_HERSHEY_COMPLEX,
                    0.5,
                    color,
                    1,
                )

            if question_assignment == None:
                assigned_answers_idx = input(
                    f"enter answer id (shown in red) to add to question{question_id}"
                )
                assigned_answers_idx = [
                    int(i.strip()) for i in assigned_answers_idx.split(",")
                ]
            else:
                assigned_answers_idx = question_assignment[question_id]

            # print(f"{question_id}:{assigned_answers_idx}")
            for idx in assigned_answers_idx:
                answer_coords = all_spots[idx]
                unassigned_answers.remove(answer_coords)
                answer = Answer(
                    x=answer_coords[0],
                    y=answer_coords[1],
                    value=f"val_{idx}",
                    circle_radius=circle_radius,
                )
                assigned_answers.append(answer)

            question_id += 1
            if question_config:
                question = Question(
                    question_id=question_id,
                    answers=assigned_answers,
                    multiple=question_config[question_id],
                )

            else:
                question = Question(question_id, assigned_answers, True)

            questions.append(question)

        template = Template(raw_img, questions, circle_radius)

        return template

    @classmethod
    def from_json(cls, json_path: str, img_path: str) -> Form:
        """Return Form instance from pre-configured JSON.

        Args:
            img_path (str): Path to image of template.
            json_path (str): Path to JSON containing configuration for form template, see format below.

        .. code-block:: json

            {"config":
                    {"radius":"<CIRCLE_RADIUS>"},
                "questions":
                    {"question_id":
                        [
                            {"answer_val":"<ANSWER_VAL>",
                            "answer_coords": ["<X_COORD>", "<Y_COORD>"]}
                        ]
                    }
                }


        Returns:
            Form: Return form instantiated from JSON config and image.
        """

        with open(json_path, "r") as fp:
            template = json.load(fp)

        return cls.from_dict(template, img_path)

    @classmethod
    def from_dict(cls, template: dict, img_path: str) -> Template:
        """Create template from dictionary of template config.

        Args:
            template (dict): template config as shown below.

        .. code-block:: python

            {"config":
                    {"radius":"<CIRCLE_RADIUS>"},
                "questions":
                    {"question_id":
                        [
                            {"answer_val":"<ANSWER_VAL>",
                            "answer_coords": ["<X_COORD>", "<Y_COORD>"]}
                        ]
                    }
                }

        img_path (str): path to image of form

        Returns:
            Template
        """
        img = cv2.imread(img_path)

        question_objs = []
        questions = template["questions"]
        question_ids = questions.keys()
        circle_radius = template["config"]["radius"]
        for question_id in question_ids:
            answers = []

            for answer in questions[question_id]:
                x, y = answer["answer_coords"]
                answer_val = answer["answer_val"]
                answers.append(Answer(x, y, answer_val, circle_radius))

            question = Question(
                question_id=int(question_id), answers=answers, multiple=None
            )
            question_objs.append(question)

        return Template(img, question_objs, circle_radius)

    def to_dict(self) -> dict:
        """Convert template obj to dictionary. See docs for dictionary structure.

        Return:
            dict : dictionary representation of template as below.

        .. code-block:: python

            {"config":
                    {"radius":"<CIRCLE_RADIUS>"},
                "questions":
                    {"question_id":
                        [
                            {"answer_val":"<ANSWER_VAL>",
                            "answer_coords": ["<X_COORD>", "<Y_COORD>"]}
                        ]
                    }
                }

        """

        question_dict = {}

        for question in self.questions:
            question_dict[question.question_id] = []

            for answer in question.answers:
                question_dict[question.question_id].append(
                    {
                        "answer_val": answer.value,
                        "answer_coords": (answer.x, answer.y),
                    }
                )
        template_dict = {}
        template_dict["config"]["radius"] = self.circle_radius
        template_dict["questions"] = question_dict
        return template_dict

    def to_json(self) -> str:
        """Convert template into json string representation.

        Returns:
            str: json string in format below.

        .. code-block:: json

            {"config":
                    {"radius":"<CIRCLE_RADIUS>"},
                "questions":
                    {"question_id":
                        [
                            {"answer_val":"<ANSWER_VAL>",
                            "answer_coords": ["<X_COORD>", "<Y_COORD>"]}
                        ]
                    }
                }

        """
        return json.dumps(self.to_dict())

    @property
    def perspective_matrix(self) -> np.ndarray:
        """return perspective matrix from outer box detected in image
        used for alignment of template

        Returns:
            np.ndarray: perspective matrix
        """
        ordered_pts = ip.get_outer_box(self.img)
        matrix = ip.get_perspective_matrix(ordered_pts)
        return matrix



from typing import Tuple

import cv2
import numpy as np


class Answer:
    """A class to represent a single answer circle"""

    def __init__(
        self,
        x: int,
        y: int,
        value: str,
        circle_radius: int,
        filled_threshold: float = 0.8,
    ):
        """Initialise an answer: individual spot corresponding to one possible answer.

        Args:
            x (int): x coordinate of answer
            y (int):  y coordinate of answer
            value (str):  value that the answer represents.
            circle_radius (int): radius of answer spot
            filled_threshold (float, optional): Value from 0.0 - 1.0 used as threshold to determine
            what fill percentage counts as a filled in answer circle. Defaults to 0.8.
        """

        self.x = x
        self.y = y
        self.circle_radius = circle_radius
        self.value = value
        self.filled_threshold = filled_threshold

    def is_filled(self, form_img: np.ndarray) -> bool:
        return self.calc_filled_perc(form_img) >= self.filled_threshold

    def mark_answer(
        self,
        form_img: np.ndarray,
        circle_thickness: int = -1,
        colour: Tuple[int] = (0, 0, 255),
    ) -> None:
        """Marks answer on image in place

        Args:
            form_img (np.ndarray): image of form read into array e.g. via cv2.imread()
            circle_thickness (int, optional): thickness of circle if a hollow circle is required.
            Defaults to -1 (filled circle).
            colour (Tuple[int], optional): colour of circle in BGR. Defaults to (0, 0, 255).

        Returns:
            None: marks answer in place so original image is modified
        """
        if len(form_img.shape) != 3:
            form_img = cv2.cvtColor(form_img, cv2.COLOR_GRAY2BGR)

        cv2.circle(
            form_img,
            (self.x, self.y),
            self.circle_radius,
            colour,
            circle_thickness,
        )

    def calc_filled_perc(self, form_img: np.ndarray) -> float:
        """Returns percentage of circle that is filled in

        Args:
            form_img (np.ndarray): _description_

        Returns:
            float: range from 0.0 - 1.0 representing percentage of circle filled in
        """
        # create full black mask with same size as image
        mask = np.zeros(form_img.shape, dtype="uint8")
        # draw white circle on mask to expose the answer circle
        cv2.circle(mask, (self.x, self.y), self.circle_radius, 255, -1)
        mask_pixels = cv2.countNonZero(mask)
        # white pixels on answer == filled in parts of answer (img colour is inverted)
        # bitwise AND == how many white pixels are on the answer
        mask = cv2.bitwise_and(form_img, mask)
        # proportion of white pixels on answer to total white pixels available in answer circle
        pct_filled = cv2.countNonZero(mask) / mask_pixels

        return pct_filled



from __future__ import annotations

import cv2
import numpy as np


class ImageAlignmentError(Exception):
    pass


def show_img(img: np.ndarray | list[np.ndarray], time: int = 0) -> None:
    """Utility to display all images in a window with window name = window(i)
    where i = index

    Args:
        img (np.ndarray | list[np.ndarray]): image to display as
        array read into array e.g. via cv2.imread()
        time (int, optional): Delay time in millisec to display window. Defaults to 0.
    """

    if type(img) != list:
        cv2.imshow("window", img)
    else:
        for i, im in enumerate(img):
            cv2.imshow("window" + str(i), im)
    cv2.waitKey(time)
    if time == 0:
        cv2.destroyAllWindows()


def thresh_img(
    img: np.ndarray, min_thresh: int = 100, max_thresh: int = 255
) -> np.ndarray:
    """Convert image to binary black and white image using thresholds

    Args:
        img (np.ndarray): image to threshold read into array e.g. via cv2.imread()
        min_thresh (int, optional): Pixels below this grayscale value will be converted to white pixels . Defaults to 100.
        max_thresh (int, optional): Pixels above this grayscale value will be converted to black pixels. Defaults to 255.

    Returns:
        np.ndarray: thresholded image with only black or white pixels
    """
    if len(img.shape) == 3 and img.shape[2] == 3:
        # three channels aka coloured img
        img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    else:
        # img already greyscale
        img_gray = img
    _, img_thresh = cv2.threshold(
        img_gray, min_thresh, max_thresh, cv2.THRESH_BINARY_INV
    )

    return img_thresh


def get_perspective_matrix(ordered_corner_pts: np.ndarray) -> np.ndarray:
    """Calculates a matrix from the corners of the aligment feature

    Args:
        ordered_corner_pts (np.ndarray): Corner points of rectangle alignment feature,
        ordered from top-left clockwise

    Returns:
        np.ndarray: perspective matrix to use to align page using cv2.getPerspectiveTransform()
    """
    # unpack ordered pts to find widths and heights
    (tl, tr, br, bl) = ordered_corner_pts

    # use euclidean distances (finally a use for pythagoras lol) - taken from pyimagesearch.com
    widthA = np.sqrt(((br[0] - bl[0]) ** 2) + ((br[1] - bl[1]) ** 2))
    widthB = np.sqrt(((tr[0] - tl[0]) ** 2) + ((tr[1] - tl[1]) ** 2))
    heightA = np.sqrt(((tr[0] - br[0]) ** 2) + ((tr[1] - br[1]) ** 2))
    heightB = np.sqrt(((tl[0] - bl[0]) ** 2) + ((tl[1] - bl[1]) ** 2))

    # find max heights/widths - i.e from top/bottom, left/right
    maxWidth = max(int(widthA), int(widthB))
    maxHeight = max(int(heightA), int(heightB))

    # initialise dst array for transformation - i.e. map corners of img onto this matrix
    dst = np.array(
        [
            [0, 0],
            [maxWidth - 1, 0],
            [maxWidth - 1, maxHeight - 1],
            [0, maxHeight - 1],
        ],
        dtype="float32",
    )

    return dst


def get_outer_box(img: np.ndarray) -> np.ndarray:
    """Finds the rectangle alignment feature in the image

    Args:
        img (np.ndarray): image to detect the outer box from

    Raises:
        ImageAlignmentError: if outer box is not detected

    Returns:
        np.ndarray: returns the 4 coordinates of the corners of the rectangle alignment feature,
        ordered from top-left clockwise
    """
    # enhance image to improve contour detection
    if len(img.shape) == 3:
        img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    else:
        img_gray = img
    img_bilat = cv2.bilateralFilter(img_gray, 11, 500, 0)
    img_edge = cv2.Canny(img_bilat, 20, 100)

    ##find outer rectangle

    conts, _ = cv2.findContours(
        img_edge, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE
    )
    # find top 10 biggest contours by area
    areas = []
    for c in conts:
        areas.append([c, cv2.contourArea(c), cv2.arcLength(c, True)])

    # sort by arclength and area - (neither work consistently for now :( )
    length_sort = sorted(areas, key=lambda x: x[2], reverse=True)
    area_sort = sorted(areas, key=lambda x: x[1], reverse=True)

    area_conts = [i[0] for i in area_sort]
    length_conts = [i[0] for i in length_sort]

    area_cont_max_x = max(i[0][0] for i in area_conts[0])
    area_cont_max_y = max(i[0][1] for i in area_conts[0])
    length_cont_max_x = max(i[0][0] for i in length_conts[0])
    length_cont_max_y = max(i[0][1] for i in length_conts[0])

    if area_cont_max_x >= length_cont_max_x:
        if area_cont_max_y >= length_cont_max_y:
            # outer contour found with area method
            top_contours = area_sort[:5]
        else:
            raise ImageAlignmentError(
                "Image Alignment Failed: outer box not detected"
            )
    elif area_cont_max_x < length_cont_max_x:
        if area_cont_max_y < length_cont_max_y:
            # outer contour found with length method
            top_contours = length_sort[:5]
        else:
            raise ImageAlignmentError(
                "Image Alignment Failed: outer box not detected"
            )

    # find outer rectangle
    for c in top_contours:
        # approximate curve to check if its rectangularish
        approx = cv2.approxPolyDP(c[0], 0.0015 * c[2], True)
        # check number of corners in apprximated polygon
        if len(approx) >= 4:
            pts = np.zeros((4, 2), dtype="float32")
            # top left
            pts[0] = approx[np.argmin(approx.sum(axis=2))]
            # top right
            pts[2] = approx[np.argmax(approx.sum(axis=2))]

            # smallest difference x-y = top right
            pts[1] = approx[np.argmin(np.diff(approx, axis=2))]
            # largest difference x-y = bottom left
            pts[3] = approx[np.argmax(np.diff(approx, axis=2))]

            break

    return pts


def align_page(
    img: np.ndarray, corner_pts: np.ndarray | None = None
) -> np.ndarray:
    """Applys perspective transform to align the image using a rectangle alignment feature on the image

    Args:
        img (np.ndarray): image of form or template read into array e.g. via cv2.imread()
        corner_pts (np.ndarray | None, optional): 4 coordinates of the corners of the rectangle alignment feature,
        ordered from top-left clockwise. Defaults to None, and the rectangle feature will be detected automatically.

    Returns:
        np.ndarray: aligned image
    """

    if corner_pts is not None:
        ordered_pts = corner_pts
    else:
        ordered_pts = get_outer_box(img)

    dst = get_perspective_matrix(ordered_pts)

    width = int(dst[2][0])
    height = int(dst[2][1])

    # transformation matrix
    matrix = cv2.getPerspectiveTransform(ordered_pts, dst)

    # transform image and resize to original size (map spots to correct locations)
    img_warp = cv2.warpPerspective(img, matrix, (width, height))

    # TODO add rotation feature (separate func probably) -
    # TODO add detection mechanism for choosing rotation e.g. top left spot or something

    # if img_warp.shape[0] > img_warp.shape[1]:
    #     img_warp = cv2.rotate(img_warp, cv2.ROTATE_90_CLOCKWISE)

    return img_warp


def process_img(img: np.ndarray) -> np.ndarray:
    """Converts image to binary black & white and aligns the page using the rectangle alignment feature

    Args:
        img (np.ndarray): image read into array e.g. via cv2.imread()

    Returns:
        np.ndarray: binary black and white, aligned image
    """

    img_thresh = thresh_img(img)
    img_aligned = align_page(img_thresh)
    return img_aligned



from __future__ import annotations

from typing import Tuple

import cv2
import numpy as np

import formpy.utils.img_processing as ip

from .template import Template


class Form:
    """A class to represent a form."""

    def __init__(self, img: np.ndarray, template: Template) -> Form:
        """Initialise form with an associated template that it was built from

        Args:
            img (np.ndarray): a form image read into array e.g. via cv2.imread()
            template (Template): template that the form was built from

        Returns:
            Form
        """
        self.template = template
        self.img = self.__resize_img(img)
        self.questions = template.questions

    def __repr__(self) -> str:
        return f"Form with {len(self.questions)} questions and {sum([len(i.answers) for i in self.questions])}"

    def __resize_img(self, img: np.ndarray) -> np.ndarray:
        """resize image to be of same size as template

        Args:
            img (np.ndarray): form image read into array e.g. via cv2.imread()
        """
        processed_img = ip.process_img(img)
        resized_img = cv2.resize(
            processed_img,
            (self.template.img.shape[1], self.template.img.shape[0]),
            interpolation=cv2.INTER_LINEAR,
        )
        return resized_img

    def mark_all_answers(self, colour: Tuple[int] = (0, 0, 255)) -> np.ndarray:
        """mark all answers on the form image with the question id and answer value

        Args:
            colour (Tuple[int], optional): color of text and circles to mark answers in BGR. Defaults to (0, 0, 255).

        Returns:
            np.ndarray: form image with marked answers, answer values and question ids
        """
        colour_img = cv2.cvtColor(
            cv2.bitwise_not(self.img), cv2.COLOR_GRAY2BGR
        )

        for qn in self.template.questions:
            answers = qn.answers
            for ans in answers:

                cv2.putText(
                    colour_img,
                    str(qn.question_id) + ans.value,
                    (ans.x, ans.y),
                    cv2.FONT_HERSHEY_COMPLEX,
                    0.4,
                    colour,
                )

                ans.mark_answer(colour_img, colour=colour, circle_thickness=-1)

        return colour_img



from __future__ import annotations

import cv2
import numpy as np


def find_spots(
    img: np.array, max_radius: int = 35, min_radius: int = 10
) -> list[list[int]]:
    """Use contour detection to find all circles that may correspond to answers on a template.
    Min and max radius and the aspect ratio of between 0.9 - 1.1 are used to determine if
    the detected contour is an answer

    Args:
        img (np.array): image of template read into array e.g. via cv2.imread()
        This should be a template page with all the answers filled in so the coordinates
        can be detected
        max_radius (int, optional): max threshold used to determine if the contour
        is a answer circle. Defaults to 35.
        min_radius (int, optional): min threshold used to determine if the contour
        is a answer circle. Defaults to 10.

    Returns:
        list[list[int]]: sorted list of [x,y] coordinates representing the centre of the answer circle.
        Sorted by x first, then y
    """
    conts, _ = cv2.findContours(img, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)

    spotCentres = []
    for c in conts:
        # compute enclosing rect to get center points
        x, y, width, height = cv2.boundingRect(c)

        aspect_ratio = width / height
        # check if bounding rectangle of circle roughly matches criteria:(aspect ratio ==1) and (w,h == 25)
        if (
            min_radius * 2 < width < max_radius * 2
            and min_radius * 2 < height < max_radius * 2
            and 0.9 < aspect_ratio < 1.1
        ):
            xCentre = int(x + (width / 2))
            yCentre = int(y + (height / 2))
            cv2.circle(img, (xCentre, yCentre), 1, 0, -1)
            spotCentres.append([xCentre, yCentre])

    # sort by x, then y
    sortedSpots = sorted(
        spotCentres, key=lambda x: (x[0], x[1]), reverse=False
    )

    return sortedSpots
